// This file is currently unlocked
//
//	Current source version is 2.5.1
//			(update this every time you change it)
//
// 2.5.1        12 May 1997     Ian Brown
//      Cryptix version string read from cryptix.security.Cryptix
//
// 2.5.0        10 May 1997     Ian Brown
//      Source control info and doccomments added
//
// If this file is unlocked, or locked by someone other than you, then you
// may download this file for incorporation into your build, however any
// changes you make to your copy of this file will be considered volatile,
// and will not be incorporated into the master build which is maintained
// at this web site.
//
// If you wish to make permanent changes to this file, you must wait until
// it becomes unlocked (if it is not unlocked already), and then lock it
// yourself. Whenever the file is locked by you, you may upload new master
// versions to the master site. Make sure you update the history information
// above. When your changes are complete, you should relinquish the lock to
// make it available for other developers.

package cryptix.pgp;

import java.io.*;
import cryptix.mime.Base64;
import cryptix.security.Cryptix;

/**
 * Converts data to and from the 64-bit encoding, with a 24-bit checksum,
 * that PGP uses as 'transport armour'.
 * <hr>
 * <P>
 * <bold>Copyright</bold> &#169 1995, 1996, 1997
 * <a href="http://www.systemics.com/">Systemics Ltd</a>
 * on behalf of the
 * <a href="http://www.systemics.com/docs/cryptix/">
 * Cryptix Development Team</a>.
 * All rights reserved.
 */
public final class Armoury
{
	/** The string (\r\n) PGP uses as a 'canonical' linefeed. */
	public static final String LINEFEED = "\r\n";

	private Armoury()
	{
	}

	/**
	 * Armour and return the data in <code>byte[]</code>
	 * <i>without</i> the PGP framing lines.
	 */
	public static String
	messageWithChecksum(byte buf[])
	{
		byte checksum[] = new byte[3];
		int csum = CRC.checksum(buf);
		checksum[0] = (byte)(csum >>> 16);
		checksum[1] = (byte)(csum >>> 8);
		checksum[2] = (byte)(csum);

		return Base64.encode(buf)+"="+Base64.encode(checksum);
	}
	
	/**
	 * Armour and write the <code>String</code> packet in <code>byte[]</code>
	 * to <code>PrintStream</code>.
	 */
	public static void
	message( PrintStream out, String name, byte buf[])
	{
		out.println( "-----BEGIN PGP "+name+"-----" );
		out.println( "Version: " + Cryptix.getVersionString() );
		out.println();
		out.print( messageWithChecksum(buf ));
		out.println( "-----END PGP "+name+"-----" );
	}
	
	/**
	 * Armour and return the <code>String</code> packet in <code>byte[]</code>.
	 */
	public static String
	message( String name, byte buf[] )
	{
		String framed;
		framed = "-----BEGIN PGP " + name + "-----" + LINEFEED;
		framed += "Version: " + Cryptix.getVersionString() + LINEFEED + LINEFEED;
		framed += messageWithChecksum(buf);
		framed += "-----END PGP " + name + "-----" + LINEFEED;
		return framed;
	}

	/**
	 * Armour and write <code>Packet</code> to <code>PrintStream</code>.
	 */
	public static void
	message( PrintStream out, Packet pkt)
	throws IOException
	{
		message(out, name(pkt), pkt.save());
	}

 	/**
 	 * 'Strips' armour off PGP binary data in <code>String</code>
 	 * and checks its 24-bit checksum.
 	 */
 	public static byte[]
	strip( String buf )
	throws FormatException, IOException, InvalidChecksumException
	{
		String checksum, armour;
		int pos;
		// Find start of ASCII armour
		pos = buf.indexOf( "-----BEGIN PGP" ) ;

		if ( pos == -1 ) throw new FormatException( "Couldn't find PGP part in message." );

		// Now get rid of version / comments
		pos = buf.indexOf( LINEFEED + LINEFEED, pos );

		// Get message up to checksum
		armour = buf.substring( pos + 4, ( pos = buf.indexOf( LINEFEED + "=" ) ) );
		checksum = buf.substring( pos + 2, pos + 7 );
		// We now have message in armour and its checksum
		int givenChecksum = decodeChecksum( checksum );
		int realChecksum = CRC.checksum( Base64.decode( armour ) );
		// Do the two checksums match?
		if ( givenChecksum != realChecksum ) throw new InvalidChecksumException( "Message checksum wrong." );

		// OK - checksum correct, so return decoded armour
		return Base64.decode( ArmouredMessage.byteEncode( armour ) );
 	}

	/**
	 * Returns the name PGP uses in its framing lines for this type
	 * of <code>Packet</code>.
	 */
	public static String
	name( Packet pkt )
	{
		if ( pkt instanceof Signature )
			return "SIGNATURE";
		else if ( pkt instanceof SecretKeyCertificate )
			return "SECRET KEY BLOCK";
		else if ( pkt instanceof PublicKeyCertificate )
			return "PUBLIC KEY BLOCK";

		throw new IllegalArgumentException("PGP Packet type "+pkt.getClass().getName()+" not valid for this method");
	}

	/**
	 * Read from <code>DataInputStream</code> until a PGP section appears.
	 */
	public static void
	readUntilBegin(DataInputStream dis)
	throws IOException
	{
		String line;
		// Read from stream until a PGP section appears
		while ( ! ( ( line = dis.readLine() ).startsWith( "-----BEGIN PGP" ) ) );
		// Now get rid of version / comments
		while ( ! ( ( line = dis.readLine() ).equals( "" ) ) );
	}

	/**
	 * Decode the checksum contained in <code>String</code>.
	 */
	public static int
	decodeChecksum(String s)
		//throws InvalidChecksumException
	{
		if (s.length() != 5)
		{
		}

		byte[] asc_sum = new byte[4];
		s.getBytes(1, 5, asc_sum, 0);
		byte[] binsum = Base64.decode(asc_sum);
		int sum = ((((int)binsum[0] & 0xFF) << 16)
			| (((int)binsum[1] & 0xFF) << 8)
			| ((int)binsum[2]) & 0xFF);
		return sum;
	}
}

