// This file is currently unlocked (change this line if you lock the file)
//
// Source code version is 2.5.7
//                        (increment this number EVERY time you make a change)
//
// History: (add to this EVERY time you make a change) 
// (insert details of your change immediately below this line)
//
// 2.5.7    11 Jul 1997         Jill Baker
//      Added method getAllKnownPublicKeys()
//
// 2.5.6    13 Jun 1997         Jill Baker
//      Added method getPrimarySecretKeyRingEntry()
//
// 2.5.5    12 Jun 1997         Jill Baker
//      support added for unencrypted secret keys.
//      Added method getSecretKey( KeyID id ).
//
// 2.5.4    11 Jun 1997         Jill Baker
//      added methods getSecretKey( String name, byte[] hash ), and
//      getSecretKey ( KeyID id, byte[] hash ).
//
//
// 2.5.3	10 May 1997         Ian Brown
//      check whether user has specified a KeyClient before checking it!
//
// 2.5.2        29 Apr 1997     Ian Brown
//      added getKey( EmailAddress ) method.
//      tidied up comments
//      changed method of adding a keyserver to setKeyServer( KeyClient )
//
// 2.5.1        15 Apr 1997     Jill Baker
//      imports moved to new namespace
//      added getKeyTrust() method.
//
//      Line temporarily commented out in setWebServer() because it won't
//      compile otherwise. This is presumably temporary until class WebClient
//      gets updated.
//
// 2.5.0        24 Mar 1997     Original Author Ian Brown
//      Original version
//
// If this file is unlocked, or locked by someone other than you, then you
// may download this file for incorporation into your build, however any
// changes you make to your copy of this file will be considered volatile,
// and will not be incorporated into the master build which is maintained
// at this web site.
//
// If you wish to make permanent changes to this file, you must wait until
// it becomes unlocked (if it is not unlocked already), and then lock it
// yourself. Whenever the file is locked by you, you may upload new master
// versions to the master site. Make sure you update the history information
// above. When your changes are complete, you should relinquish the lock to
// make it available for other developers.

package cryptix.pgp;

import java.io.IOException;
import java.net.UnknownHostException;
import java.util.Vector;
import cryptix.security.rsa.PublicKey;
import cryptix.security.rsa.SecretKey;
import cryptix.util.EmailAddress;

/**
 * Stores all of the key sources known to a program - 
 * all public and secret keyrings, and a Web master keyserver.
 * If setKeyServer is not used, the object will not use one.<p>
 * Can be passed to objects which decrypt and verify PGP messages
 * such as ArmouredMessage.<p>
 * NB Keyrings are searched in the order in which they are added.
 * Therefore, <b>always add the most highly trusted keyrings first.</b><p>
 * The master keyserver is only consulted if a key is not available locally.<p> 
 * <hr>
 * <P>
 * <bold>Copyright</bold> &#169 1995, 1996, 1997
 * <a href="http://www.systemics.com/">Systemics Ltd</a>
 * on behalf of the
 * <a href="http://www.systemics.com/docs/cryptix/">
 * Cryptix Development Team</a>.
 * All rights reserved.
 * @author Ian Brown
 * @author Jill Baker
 */
public final class KeyGlob extends KeyStore
{

	private Vector publicKeyRings;
	private Vector secretKeyRings;
	private KeyClient keyClient;
	
	/**
	 * Initialise an empty KeyGlob
	 * Use addKeyRing and setKeyServer to fill it up
	 */
	public KeyGlob()
	{
		publicKeyRings = new Vector();
		secretKeyRings = new Vector();
		keyClient = null;
	}

	/**
	 * Add publicKeyRing to the KeyGlob's collection
	 */
	public void
	addKeyRing( PublicKeyRing publicKeyRing )
	{

		publicKeyRings.addElement( publicKeyRing );
	}

	/**
	 * Add secretKeyRing to the KeyGlob's collection.
	 * By allowing more than one secret keyring, users can keep a standard secret key available,
	 * and have a high-security key stored on a floppy disk which applications only access
	 * transparently through this class when it's needed.
	 */
	public void
	addKeyRing( SecretKeyRing secretKeyRing )
	{
		secretKeyRings.addElement( secretKeyRing );
	}
	
	/**
	 * Give KeyGlob keyclient object it can use to fetch keys remotely.
	 * This will be searched for public keys if they are not available locally.
	 * @param client An instantiated KeyClient subclass object.
	 */
	public void
	setKeyServer( KeyClient client )
	throws IOException, UnknownHostException
	{
		keyClient = client;
	}
	
	/**
         * Get the trust value for a given key ID
         * @param keyID the key ID you are interested in
         * @param publicRings search public rings (true), or secret rings (false)
         * @return trust value, or -1 if key not present
         */
        public int getKeyTrust( KeyID keyID, boolean publicRings )
        {

            int trust;

            // Cycle through all public keyrings to look for key ID
            if (publicRings)
            {
                for ( int i = 0; i < publicKeyRings.size(); i++ )
		{
                    PublicKeyRing ring = (PublicKeyRing) publicKeyRings.elementAt( i );
                    trust = ring.getKeyTrust( keyID );
                    if (trust != -1) return trust;
		}
            }
            else
            {
                for ( int i = 0; i < secretKeyRings.size(); i++ )
		{
                    SecretKeyRing ring = (SecretKeyRing) secretKeyRings.elementAt( i );
                    trust = ring.getKeyTrust( keyID );
                    if (trust != -1) return trust;
		}
            }
            return -1;
        }

	public KeyRingEntry
	getPublicKeyRingEntry( String name )
	throws IOException
	{
		KeyRingEntry entry = null;

		// Cycle through all public keyrings to look for name
		for ( int i = 0; i < publicKeyRings.size(); i++ )
		{
			entry = ( (PublicKeyRing) publicKeyRings.elementAt( i ) ).getKeyRingEntry( name );
			// If we've found it, return immediately
			if ( entry != null )
				return entry;
		}

		// If user has specified a keyserver...
		if ( keyClient != null )
			// then try it for the key
			return keyClient.getKeyRingEntry( name );
		else return null;
	}
	
	/**
	 * Try and find a public key for name.
	 * @param name The e-mail address of the user
	 * @return Their public key, or null if one can't be found
	 */
	public PublicKey
	getKey( String name )
	throws IOException
	{
		KeyRingEntry entry = getPublicKeyRingEntry( name );
		if ( entry != null ) return entry.publicKey();
		else return null;
	}

	/**
	 * Try and find a public keyring entry for id.
	 * @param name The key ID of the user (the low 8 bits of their key pair's public modulus)
	 * @return Their public keyring entry, or null if one can't be found
	 */
	public KeyRingEntry
	getPublicKeyRingEntry ( KeyID id )
	throws IOException
	{
		KeyRingEntry entry = null;

		// Cycle through all public keyrings to look for id
		for ( int i = 0; i < publicKeyRings.size(); i++ )
		{
			entry = ( (PublicKeyRing) publicKeyRings.elementAt( i ) ).getKeyRingEntry( id );
			// If we've found it, return immediately
			if ( entry != null )
				return entry;
		}

		// If user has specified a keyserver...
		if ( keyClient != null )
			// then try it for the key
			return keyClient.getKeyRingEntry( id );
		else return null;
	}

	/**
	 * Try and find a public key for name.
	 * @param name The e-mail address of the user
	 * @return Their public key, or null if one can't be found
	 */
	public PublicKey
	getKey( KeyID id )
	throws IOException
	{
		KeyRingEntry entry = getPublicKeyRingEntry( id );
		if ( entry != null ) return entry.publicKey();
		else return null;
	}
		
	/**
	 * Try and find a public keyring entry for address
	 * @param address The e-mail address of the user
	 * @return Their public keyring entry, or null if one can't be found
	 */
	public final KeyRingEntry
	getPublicKeyRingEntry ( EmailAddress address )
	throws IOException
	{
		KeyRingEntry entry = null;

		// Cycle through all public keyrings to look for address
		for ( int i = 0; i < publicKeyRings.size(); i++ )
		{
			entry = ( (PublicKeyRing) publicKeyRings.elementAt( i ) ).getKeyRingEntry( address );
			// If we've found it, return immediately
			if ( entry != null )
				return entry;
		}
		
		// If user has specified a keyserver...
		if ( keyClient != null )
			// then try it for the key
			return keyClient.getKeyRingEntry( address );
		else return null;
	}
	

	/**
	 * Try and find a public key for address.
	 * @param name The e-mail address of the user
	 * @return Their public key, or null if one can't be found
	 */
	public PublicKey
	getKey( EmailAddress address )
	throws IOException
	{
		KeyRingEntry entry = getPublicKeyRingEntry( address );
		if ( entry != null ) return entry.publicKey();
		else return null;
	}	

	/**
	 * Get the first KeyRingEntry available on the secret keyring
	 * @return Their secret keyring entry, or null if one can't be found
	 */
	public KeyRingEntry getPrimarySecretKeyRingEntry() throws IOException
	{
		SecretKeyRing keyRing = (SecretKeyRing) secretKeyRings.elementAt( 0 );
        return keyRing.getPrimaryKeyRingEntry(); 
	}

	/**
	 * Try and find a secret keyring entry for name
	 * @param name The user ID of the user
	 * @return Their secret keyring entry, or null if one can't be found
	 */
	public KeyRingEntry
	getSecretKeyRingEntry( String name )
	throws IOException
	{
		KeyRingEntry entry = null;
		// Cycle through all secret keyrings to look for name
		for ( int i = 0; i < secretKeyRings.size(); i++ )
		{
			entry = ( (SecretKeyRing) secretKeyRings.elementAt( i ) ).getKeyRingEntry( name );
			// If we've found it, return immediately
			if ( entry != null )
				return entry;
		}
		// It's not in any of them
		return null;
	}

	/**
	 * Try and find a secret key for name and decrypt it with passphrase
	 * @param name The user ID of the user
	 * @param passphrase The passphrase protecting their secret key
	 * @return Their secret key, or null if one can't be found
	 */
	public SecretKey
	getSecretKey( String name, Passphrase passphrase )
	throws IOException, DecryptException
	{
		KeyRingEntry entry = getSecretKeyRingEntry( name );
		if ( entry != null )
		{
			SecretKeyCertificate cert = (SecretKeyCertificate)( entry.certificate() );
			return cert.getSecretKey( passphrase );
		}
		else return null;
	}

	/**
	 * Try and find a secret key for name and decrypt it with hash
	 * @param name The user ID of the user
	 * @param hash The hash protecting their secret key
	 * @return Their secret key, or null if one can't be found
	 */
	public SecretKey
	getSecretKey( String name, byte[] hash )
	throws IOException, DecryptException
	{
		KeyRingEntry entry = getSecretKeyRingEntry( name );
		if ( entry != null )
		{
			SecretKeyCertificate cert = (SecretKeyCertificate)( entry.certificate() );
			return cert.getSecretKey( hash );
		}
		else return null;
	}

	/**
	 * Try and find a secret keyring entry for id
	 * @param name The key ID of the user's key
	 * @return Their secret keyring entry, or null if one can't be found
	 */
	public KeyRingEntry
	getSecretKeyRingEntry( KeyID id )
	throws IOException
	{
		KeyRingEntry entry = null;
		// Cycle through all secret keyrings to look for id
		for ( int i = 0; i < secretKeyRings.size(); i++ )
		{
			entry = ( (SecretKeyRing) secretKeyRings.elementAt( i ) ).getKeyRingEntry( id );
			// If we've found it, return immediately
			if ( entry != null )
				return entry;
		}
		// It's not in any of them
		return null;
	}
	
	/**
	 * Try and find an unencrypted secret key for id
	 * @param id The key ID of the user's key
	 * @return Their secret key, or null if one can't be found
	 */
	public SecretKey getSecretKey ( KeyID id ) throws IOException, DecryptException
	{
		KeyRingEntry entry = getSecretKeyRingEntry( id );
		if ( entry != null )
		{
			SecretKeyCertificate cert = (SecretKeyCertificate)( entry.certificate() );
			return cert.getSecretKey();
		}
		else return null;
	}
	
	/**
	 * Try and find a secret key for id and decrypt it with passphrase
	 * @param id The key ID of the user's key
	 * @param passphrase The passphrase protecting their secret key
	 * @return Their secret key, or null if one can't be found
	 */
	public SecretKey getSecretKey ( KeyID id, Passphrase passphrase )
	throws IOException, DecryptException
	{
		KeyRingEntry entry = getSecretKeyRingEntry( id );
		if ( entry != null )
		{
			SecretKeyCertificate cert = (SecretKeyCertificate)( entry.certificate() );
			return cert.getSecretKey( passphrase );
		}
		else return null;
	}
	
	/**
	 * Try and find a secret key for id and decrypt it with hash
	 * @param name The key ID of the user's key
	 * @param hash The hash protecting their secret key
	 * @return Their secret key, or null if one can't be found
	 */
	public SecretKey
	getSecretKey ( KeyID id, byte[] hash )
	throws IOException, DecryptException
	{
		KeyRingEntry entry = getSecretKeyRingEntry( id );
		if ( entry != null )
		{
			SecretKeyCertificate cert = (SecretKeyCertificate)( entry.certificate() );
			return cert.getSecretKey( hash );
		}
		else return null;
	}
	
	/**
	 * Returns the name of a user whose key has ID keyID.<p>
	 * Only checks local keyrings.
	 */
	public final String
	getUserName( KeyID keyID )
	{
		String name = null;
		// Cycle through all public keyrings to look for name
		for ( int i = 0; i < publicKeyRings.size(); i++ )
		{
			name = ( (PublicKeyRing) publicKeyRings.elementAt( i ) ).getUserName( keyID );
			// If we've found it, return immediately
			if ( name != null )
				return name;
		}

		// We could look it up from the keyserver, but that would be overkill
		// so just follow PGP's example and give the key ID as a string
		return keyID.toString();
	}	

	public final String[] getFullUserNameList()
	{
        int count = 0;
        int offset = 0;
        Vector v = new Vector();

        // Cycle through all public keyrings
		for ( int i = 0; i < publicKeyRings.size(); i++ )
		{
			PublicKeyRing ring = ( (PublicKeyRing) publicKeyRings.elementAt( i ) );
            String[] s = ring.getFullUserNameList();
            count += s.length;
            v.addElement( s );
		}

        String[] fullList = new String[ count ];

        // Cycle through vector
		for ( int i = 0; i < v.size(); i++ )
		{
            String[] s = (String[]) v.elementAt(i);
            int len = s.length;
            System.arraycopy( s, 0, fullList, offset, len );
            offset += len;
		}

        return fullList;
	}
	
}
