// This file is currently unlocked (change this line if you lock the file)
//
// Source code version is 2.5.8
//                        (increment this number EVERY time you make a change)
//
// History: (add to this EVERY time you make a change) 
// (insert details of your change immediately below this line)
//
// 2.5.8	7 Nov 1997         Ian Brown
//		allows certificates with no user IDs - essential for authorisation SPKI-type
//		keys
//		index keys correctly ie first read is used
//
// 2.5.7	1 July 1997         Ian Brown
//	    added functions to store filename and write out keyring if it isn't dirty
//
// 2.5.6	13 Jun 1997         Jill Baker
//      added method getPrimaryKeyRingEntry(). Bug fix to method write(), which
//      previously did not preserve the order of keys in a keyring.
//      getKeyRingEntry( String username ) is now much clever.
//
// 2.5.5 	12 Jun 1997         Jill Baker
//      added method getSecretKey( KeyID )
//      (ie. support for unencrypted secret keys).
//
// 2.5.4	7 June 1997         Ian Brown
//	    added functions to return whole KeyRingEntries as well as PublicKeys
//
// 2.5.3        06 May 1997     Ian Brown
//      removed final from getKeyTrust to change its behaviour in SecretKeyRing.
//
// 2.5.2        03 May 1997     Ian Brown
//      added full doc comments.
//
// 2.5.1        01 Apr 1997     Jill Baker
//      added getKeyTrust() method.
//      Changed imports to new namespace
//
// 2.5.0        24 Mar 1997     Original Author not stated
//      Original version
//
// If this file is unlocked, or locked by someone other than you, then you
// may download this file for incorporation into your build, however any
// changes you make to your copy of this file will be considered volatile,
// and will not be incorporated into the master build which is maintained
// at this web site.
//
// If you wish to make permanent changes to this file, you must wait until
// it becomes unlocked (if it is not unlocked already), and then lock it
// yourself. Whenever the file is locked by you, you may upload new master
// versions to the master site. Make sure you update the history information
// above. When your changes are complete, you should relinquish the lock to
// make it available for other developers.

/*
 * Copyright (C) 1995, 1996, 1997 Systemics Ltd
 * on behalf of the Cryptix development team.  All rights reserved.
 */

package cryptix.pgp;

import java.io.*;
import java.util.Hashtable;
import java.util.Vector;
import java.util.Enumeration;

import cryptix.security.rsa.PublicKey;
import cryptix.security.rsa.SecretKey;

import cryptix.util.EmailAddress;

/**
 * KeyRing is an abstract superclass for keyring objects -
 * stores of keys (public and private) held in a file.
 * PublicKeyRing, KeyCache and SecretKeyRing implement
 * the specific behaviour of these different kinds of keyring files.<p>
 *
 * Note that a keyring is not automatically written out when it is destroyed.
 * The application must call <code>write()</b> itself if it has added keys
 * which is wishes to keep.<p>
 *
 * KeyRing is unlikely to be useful for any other subclasses. If you wish to
 * implement a general key retrieval class, use <code>KeyStore</code> instead.
 * If you wish to add file-specific behaviour, use PublicKeyRing or SecretKeyRing.<p>
 *
 * A <code>remove()</code> function may also be added.<p>
 *
 * <hr>
 * <P>
 * <bold>Copyright</bold> &#169 1995, 1996, 1997
 * <a href="http://www.systemics.com/">Systemics Ltd</a>
 * on behalf of the
 * <a href="http://www.systemics.com/docs/cryptix/">
 * Cryptix Development Team</a>.
 * All rights reserved.
 * @version 1.3, 7/2/1997
 * @author Systemics Ltd
 * @author Ian Brown
 * @author Jill Baker
 * @see PublicKeyRing
 * @see KeyCache 
 * @see SecretKeyRing
 */
public abstract class KeyRing extends KeyStore
{
	/** maps username to keyID */
	protected Hashtable name2keyId = new Hashtable();

	/** maps e-mail address to KeyID */
	protected Hashtable addr2keyId = new Hashtable();

   	/** maps KeyID to KeyRingEntry */
	protected Hashtable keyId2keyRingEntry = new Hashtable();

	/** this preserves the key order in the keyring */
	protected Vector keyRingEntries = new Vector();

	/** Stores the name of the file the keyring was loaded from */
	protected File filename;
	
	/** Has the keyring been altered since it was last written out */
	private boolean dirty = false;
	
	protected KeyRing(){};

	protected KeyRing( InputStream in )
	throws IOException
	{
		init( new PacketInputStream( in ) );
	}

	/**
	 * @deprecated Use KeyRing( File name ) instead.
	 */
	protected KeyRing( String filename )
	throws FileNotFoundException, IOException
	{
		FileInputStream keyfile = new FileInputStream( filename );
		init( new PacketInputStream( keyfile ) );
        	keyfile.close();
        	
        	this.filename = new File( filename );
	}

	/**
	 * Load a keyring from the file specified in <code>name</code>.
	 */
	protected KeyRing( File name )
	throws FileNotFoundException, IOException
	{
		FileInputStream keyfile = new FileInputStream( name );
		init( new PacketInputStream( keyfile ) );
        	keyfile.close();
        	
        	filename = name;
	}
	
	/**
	 * Returns the <b>primary</b> name of the user who owns
	 * the key referenced by keyID.<p>
	 * If their key is not available, the keyID is returned as a String.
	 * This mimics PGP's behaviour.
	 */
	public final String
	getUserName( KeyID keyID )
	{
		KeyRingEntry entry =  (KeyRingEntry)keyId2keyRingEntry.get( keyID );
		if ( entry != null ) return entry.primaryName();
		else return "Key: " + keyID.toString();
	}

	/**
	 * Adds a key to this keyring.<p>
	 * You can't just add a Certificate - you need at least
	 * a UserId and KeyRingTrust packet, and possibly
	 * signatures and further UserIds as well.<p>
	 * KeyRingEntry contains all of this.
	 */
	public final void
	add( KeyRingEntry entry)
	throws FormatException
	{
		KeyID id;
		String username;

       	keyRingEntries.addElement( entry );
		Certificate cert = entry.certificate();
		
		if ( cert instanceof PublicKeyCertificate )
			id = new KeyID( ((PublicKeyCertificate)cert).publicKey().id() );
		else
			id = ((SecretKeyCertificate)cert).id();

		// If a keyring is malformed and has more than one entry with the same
		// ID, the last one read will be indexed under that ID. 
		// This should not arise as PGP will not add keys to keyrings in this
		// way - the correct behaviour when adding a key with an ID the
		// same as a pre-existing one is to merge the two. Although this isn't
		// good with non-unique IDs, which is what Open PGP is proposing!
		keyId2keyRingEntry.put( id, entry );

		UserIdCertificate[] idcerts = entry.getUserIdCertificates();

		int len = idcerts.length;

		for ( int i = 0; i < len; i++ )
		{
			username = idcerts[i].userName();
			// PGP adds new user IDs FIRST in the keyring, and finds these first
			// so we need to mimic this behaviour.
			// If we don't already have a key indexed for this username
			if ( name2keyId.get( username ) == null ) // then index it
				name2keyId.put( username, id );

			if ( username.indexOf( "@")  != -1  ) // If username contains an address
			{
				try
				{
					EmailAddress address = new EmailAddress( username );
					// Ditto for address
					if ( addr2keyId.get( address ) == null )
						addr2keyId.put( address, id );
				}
				
				catch ( IOException e ) {} // username did not contain a valid address
			}
		}
		
		dirty = true;
	}

	/**
	 * Get a list of all the names on the key with ID keyID.
	 */
	public final String[]
	getFullNameList( KeyID keyID )
	{
		int len = name2keyId.size();
		if ( len > 0 )
		{
			Vector v = new Vector();
			Enumeration e = name2keyId.keys();
			for ( int i = 0; i < len; i++ )
			{
				Object o;
				if ( keyID.equals( name2keyId.get( o = e.nextElement() ) ) )
					v.addElement( o );
			}
			if ( (len = v.size()) > 0 )
			{
				String[] ret;
				v.copyInto( ret = new String[len] );
				return ret;
			}
		}
		return new String[0];
	}

	/**
	 * Get a list of all the usernames of keys on this keyring.
	 */
	public final String[]
	getFullUserNameList()
	{
		int len = name2keyId.size();
		String ret[] = new String[len];
		if ( len > 0 )
		{
			Enumeration e = name2keyId.keys();
			for ( int i = 0; i < len; i++ )
				ret[i] = (String)e.nextElement();
		}
		return ret;
	}

	protected abstract void init( PacketInputStream reader ) throws IOException;

	public final KeyRingEntry getPrimaryKeyRingEntry()
	{
        	if (keyRingEntries.size() == 0) return null;
        	return (KeyRingEntry) keyRingEntries.elementAt( 0 );
	}

	public final KeyRingEntry getKeyRingEntry( KeyID id )
	{
		return (KeyRingEntry)keyId2keyRingEntry.get( id );
	}
	
	public final KeyRingEntry getKeyRingEntry( EmailAddress address )
	{
		KeyID id = (KeyID)( addr2keyId.get( address ) );
		if ( id != null ) return getKeyRingEntry( id );
		else return null;
	}
	
	public final KeyRingEntry getKeyRingEntry( String lookFor )
	{
        	Enumeration e;
        	KeyRingEntry kre;
       		SecretKeyCertificate cert;
        	String kreName;
        	UserIdCertificate[] idcs;

        	if (lookFor.startsWith( "0x" ))
        	{
            		lookFor = lookFor.substring( 2 );

            		e = keyRingEntries.elements();
            		while (e.hasMoreElements())
            		{
               			kre = (KeyRingEntry) e.nextElement();
                		cert = (SecretKeyCertificate)kre.certificate();
                		kreName = cert.id().toString();
                		if (kreName.endsWith( lookFor )) return kre;
            		}
            		return null;
        	}
        	else
        	{
            		lookFor = lookFor.toUpperCase();

            		// First of all, see if there is an exact match
		    	KeyID id = (KeyID)( name2keyId.get( lookFor ));
		    	if (id != null)
                	return getKeyRingEntry( id );

            		// OK, no exact match, so look for a "starts with" match
            		e = keyRingEntries.elements();
            		while (e.hasMoreElements())
            		{
                		kre = (KeyRingEntry) e.nextElement();
                		idcs = kre.getUserIdCertificates();

                		for (int i=0; i<idcs.length; i++)
                		{
                    			kreName = idcs[i].userName().toUpperCase();
                   			if (kreName.startsWith( lookFor )) return kre;
                		}
            		}

            		// Last try, look for a "contains" match
            		e = keyRingEntries.elements();
            		while (e.hasMoreElements())
            		{
                		kre = (KeyRingEntry) e.nextElement();
                		idcs = kre.getUserIdCertificates();

                		for (int i=0; i<idcs.length; i++)
                		{
                    			kreName = idcs[i].userName().toUpperCase();
                    			if (kreName.indexOf( lookFor ) != -1) return kre;
                		}
            		}

            		// It isn't there. Give up.
            		return null;
        	}
	}
		
	protected final PublicKey
	getPublicKey( KeyID id )
	{
		KeyRingEntry entry;
		if ( ( entry = getKeyRingEntry( id ) ) != null )
		{
			Certificate c;
			if ( ( c = entry.certificate() ) instanceof PublicKeyCertificate )
				return ((PublicKeyCertificate)c).publicKey();
		}
		return null;
	}

	protected final PublicKey
	getPublicKey( EmailAddress address )
	{
		KeyRingEntry entry;
		KeyID id = (KeyID)addr2keyId.get( address );
		if ( id == null ) return null;
		else if ( (entry = (KeyRingEntry)keyId2keyRingEntry.get( id ) ) != null )
		{
			Certificate c;
			if ( (c = entry.certificate()) instanceof PublicKeyCertificate )
				return ((PublicKeyCertificate)c).publicKey();
		}
		return null;
	}

    	protected final SecretKey getSecretKey( KeyID id ) throws DecryptException
    	{
		KeyRingEntry entry;
		if ( (entry = (KeyRingEntry)keyId2keyRingEntry.get( id )) != null )
		{
			Certificate c;
			if ( (c = entry.certificate()) instanceof SecretKeyCertificate )
				return ((SecretKeyCertificate)c).getSecretKey();
		}
		return null;
    	}

    	protected final SecretKey
	getSecretKey( KeyID id, Passphrase passphrase )
	throws DecryptException
	{
		KeyRingEntry entry;
		if ( (entry = (KeyRingEntry)keyId2keyRingEntry.get( id )) != null )
		{
			Certificate c;
			if ( (c = entry.certificate()) instanceof SecretKeyCertificate )
				return ((SecretKeyCertificate)c).getSecretKey( passphrase );
		}
		return null;
	}

	protected final SecretKey
	getSecretKey( EmailAddress address, Passphrase passphrase )
	throws DecryptException
	{
		KeyRingEntry entry;
		KeyID id = (KeyID)addr2keyId.get( address );
		if ( id == null ) return null;
		else if ( (entry = (KeyRingEntry)keyId2keyRingEntry.get( id ) ) != null )
		{
			Certificate c;
			if ( (c = entry.certificate()) instanceof SecretKeyCertificate )
				return ((SecretKeyCertificate)c).getSecretKey( passphrase );
		}
		return null;
	}

	protected final PublicKey
	getIfBegins( Enumeration e, String username )
	throws IOException
	{
		while( e.hasMoreElements() )
		{
			String key = (String)e.nextElement();
			// Looks up key indirectly using keyID as a 'pointer'
			// Could this be if ( key.indexOf( username ) != -1 )
			// to make a getIfContains?
			if ( key.startsWith( username ) )
				return getKey( key );
		}
		return null;
	}

	/**
	 * Writes out all KeyRingEntries contained in this keyring
	 * to outputStream.
	 */
	public void
	write( OutputStream outputStream )
	throws IOException
	{
		PacketOutputStream pos;
		if ( outputStream instanceof PacketOutputStream) pos = (PacketOutputStream)outputStream;
		else pos = new PacketOutputStream( outputStream );
		KeyID id;
		for ( Enumeration e = keyId2keyRingEntry.keys(); e.hasMoreElements(); )
		{
			id = (KeyID)e.nextElement();
			pos.writeKeyRingEntry( (KeyRingEntry)keyId2keyRingEntry.get( id ) );
		}
	}
	
	/**
	 * If this keyring was loaded from a file (ie <code>filename</code> is set)
	 * then write it back to that file, but only if the keyring has been changed
	 * since it was last written out.
	 */
	 public void
	 write()
	 throws IOException
	 {
	 	if ( dirty )
	 	{
	 		write( new FileOutputStream( filename ) );
	 		dirty = false;
	 	}
	 }

	/** this will check the signatures assuming all the keys are in this ring */
	public boolean
	checkSignatures()
	{
		return checkSignatures( this );
	}

	/** this will check the signature assuming that the KeyStore contains all the nec keys */
	public boolean
	checkSignatures( KeyStore store )
	{
		for ( Enumeration e = keyId2keyRingEntry.elements(); e.hasMoreElements(); )
		{
			if ( !((KeyRingEntry)e.nextElement()).checkSignatures( store ) )
				return false;
		}
		return true;
	}

	/** debug dump of key ring */
	public final void
	dump( PrintStream out )
	{
		Hashtable h = name2keyId;
		out.println( "Username to KeyID mappings ..." );
		for ( Enumeration e = h.keys(); e.hasMoreElements(); )
		{
			Object o = e.nextElement();
			out.println( o + " maps to " + h.get( o ) );
		}

		out.println();
		out.println( "KeyID to Certificate/Key mappings ..." );
		h = keyId2keyRingEntry;
		for ( Enumeration e = h.keys(); e.hasMoreElements(); )
		{
			Object o = e.nextElement();
			out.println( o + " maps to " + h.get( o ) );
		}
		out.println( checkSignatures() ? "Signatures O.K." : "Signatures BAD" );
	}

	/** debug list of signatures etc */
	public final void
	list( PrintStream out )
	{
		out.println();
		out.println( "Key signature checking ..." );
		for ( Enumeration e = keyId2keyRingEntry.keys(); e.hasMoreElements(); )
		{
			KeyID id = (KeyID)e.nextElement();
			out.println( id );
			KeyRingEntry entry = (KeyRingEntry)keyId2keyRingEntry.get( id );
			UserIdCertificate[] names = entry.getUserIdCertificates();
			Certificate cert = entry.certificate();
			out.println( "Cert:" + cert );
			for( int i = 0; i < names.length; i++ )
			{
				UserId name;
				out.println( "User name:" + (name = names[i].userId()).toString() );
				Signature[] sigs = names[i].signatures();
				for ( int j = 0; j < sigs.length; j++ )
				{
					KeyID sigid = sigs[j].getKeyId();
					out.println( "Signed by " + sigid + " User name " + getUserName( sigid ) );
					try
					{
						out.println( checkSignature( sigs[j], cert, name ) ? "GOOD signature" : "*** BAD signature ***" );
					}
					catch( IOException ioe )
					{
						out.println( "Signature unchecked " + ioe.getClass().getName() + ":" + ioe.getMessage() );
					}
				}
			}
		}
		return;
	}

	/**
	 * How much is the key with ID keyId trusted by this user?
	 */
        public int
        getKeyTrust( KeyID keyId )
        {
            KeyRingEntry kre = (KeyRingEntry) keyId2keyRingEntry.get( keyId );
            if (kre == null) return -1;
            UserIdCertificate[] uidcert = kre.getUserIdCertificates();
            return uidcert[0].getTrustValue();
        }
}

