// This file is currently unlocked (change this line if you lock the file)
//
// Source code version is 2.5.10
//                        (increment this number EVERY time you make a change)
//
// History: (add to this EVERY time you make a change) 
// (insert details of your change immediately below this line)
//
// 2.5.10       16 Jun 1997     Jill Baker
//      Added new constructors (String phrase, String charset) and
//      (String phrase, String charset, long timeout ).
//
// 2.5.9        12 Jun 1997     Jill Baker
//      Bug fix in 2.5.8 didn't fix everything, so try again - the
//      PassphraseThread is now a Daemon thread. It all works now!
//  
// 2.5.8        11 Jun 1997     Jill Baker
//      Bug fix. Bad reference loop prevented garbage collection in some
//      circumstances. Now fixed, thanks to new class SensitiveByteArray.
//
//      Also, there is now only one way of getting a passphrase hash out of
//      this class, and that is using the method hash(). This avoids the
//      need for manual synchronisation by the caller.
//
// 2.5.7        03 Jun 1997     Jill Baker
//      Timeout mechanism now uses separate thread. Method expired()
//      has been deprecated. Override methods hashCode(), equals()
//      and finalize() have been added. New method expire() added.
//
// 2.5.6        27 May 1997     Jill Baker
//      Simplified ETERNITY default for timeout. 
//
// 2.5.5        23 May 1997     Jill Baker
//      Timer mechanism is now per-Passphrase. Added new constructor
//      (passphrase, timeout). Added new method expired(). md5hash()
//      now throws exception if passphrase expired.
//
// 2.5.4	    22 May 1997     Ian Brown
//	    Deprecated cipher() and hash(), added md5hash() method
//
// 2.5.3        22 May 1997     Jill Baker
//      Deleted constructors which input passphrase.
//      Use cryptix.ui.UserInterface.getPassphrase() instead.
//      Added new timer methods touch() and getTimeInterval().
//
// 2.5.2        15 Apr 1997     Jill Baker
//      Moved imports to new namespace
//
// 2.5.1        01 Apr 1997     Jill Baker
//      Can now handle alternative block cipher algorithms, not just IDEA.
//
// 2.5.0        24 Mar 1997     Original Author not stated
//      Original version
//
// If this file is unlocked, or locked by someone other than you, then you
// may download this file for incorporation into your build, however any
// changes you make to your copy of this file will be considered volatile,
// and will not be incorporated into the master build which is maintained
// at this web site.
//
// If you wish to make permanent changes to this file, you must wait until
// it becomes unlocked (if it is not unlocked already), and then lock it
// yourself. Whenever the file is locked by you, you may upload new master
// versions to the master site. Make sure you update the history information
// above. When your changes are complete, you should relinquish the lock to
// make it available for other developers.

/**
 * This class stores a passphrase's MD5 hash, which is all is needed.
 * A determined attacker could still find this hash - so keep your computer
 * physically secure and only install trusted applications.
 * @author Ian Brown
 * @author Jill Baker
 *
 * <hr>
 * <P>
 * <bold>Copyright</bold> &#169 1995, 1996, 1997
 * <a href="http://www.systemics.com/">Systemics Ltd</a>
 * on behalf of the
 * <a href="http://www.systemics.com/docs/cryptix/">
 * Cryptix Development Team</a>.
 * All rights reserved.
 */
package cryptix.pgp;

import cryptix.mime.LegacyString;
import cryptix.pgp.charset.CharSet;
import cryptix.util.SensitiveByteArray;

import cryptix.security.BlockCipher;
import cryptix.security.Blowfish;
import cryptix.security.IDEA;
import cryptix.security.MD5;

/**
 * This class represents a passphrase, which may optionally have a timer.
 * Internally, the class stores only the MD5 hash of the passphrase, not
 * the passphrase itself. In the case of timed passphrases, this hash is
 * burned (overwritten with zeroes) at the end of the timeout interval. The
 * method <samp>hash()</samp> gives you a <i>copy</i> of the hash, not the
 * original, so that your copy will still be valid if the original is burned
 * (because the timer expired before you looked at your copy). Nonetheless,
 * your copy is "safe", because it is returned as a SensitiveByteArray, so
 * it will get burned after it goes out of scope, at the next garbage
 * collect, even it you don't explicitly burn it yourself.<p>
 */
public final class Passphrase
{
    public static final long ETERNITY = 0;

	private PassphraseThread thread = null;
    private SensitiveByteArray hashSBA;

    /**
     * Constructor
     * @param phrase String to convert to Passphrase.
     *     You should destroy the String as soon as you have called this routine.
     */
    public Passphrase( String phrase )
    {
        hashSBA = new SensitiveByteArray( MD5.hash( phrase ), false );
    }

    /**
     * Constructor
     * @param phrase String to convert to Passphrase.
     *     You should destroy the String as soon as you have called this routine.
     * @param charset name of charset in which phrase String is written.
     */
    public Passphrase( String phrase, String charset )
    {
        SensitiveByteArray phraseExt = new SensitiveByteArray( LegacyString.toByteArray( phrase ), false );
        SensitiveByteArray phraseInt;
        try
        {
            phraseInt = new SensitiveByteArray( CharSet.decode( phraseExt.getData(), charset ), false );
        }
        catch (FormatException e)
        {
            phraseInt = phraseExt;
        }
        hashSBA = new SensitiveByteArray( MD5.hash( phraseInt.getData() ), true );
        phraseInt.burn();
        phraseExt.burn();
    }

    /**
     * Constructor
     * @param phrase String to convert to Passphrase.
     *     You should destroy the String as soon as you have called this routine.
     * @param timeout lifespan of Passphrase in millseconds.
     *     You can use the constant Passphrase.ETERNITY if you want.
     */
    public Passphrase( String phrase, long timeout )
    {
        this( phrase );
		if (timeout != ETERNITY) thread = new PassphraseThread( hashSBA, timeout );
    }

    /**
     * Constructor
     * @param phrase String to convert to Passphrase.
     *     You should destroy the String as soon as you have called this routine.
     * @param timeout lifespan of Passphrase in millseconds.
     *     You can use the constant Passphrase.ETERNITY if you want.
     */
    public Passphrase( String phrase, String charset, long timeout )
    {
        this( phrase, charset );
		if (timeout != ETERNITY) thread = new PassphraseThread( hashSBA, timeout );
    }

    /**
     * Reset the timer.
     */
    public synchronized void touch()
    {
		if (thread != null) thread.touch();
    }

    /**
     * Get the md5 hash of the passphrase.<p>
     *
     * @return the Passphrase as a copy of the MD5 hash contained in a SensitiveByteArray.
     * @exception cryptix.pgp.PassphraseExpiredException will be thrown if the password has expired.
     */
    public synchronized SensitiveByteArray hash() throws PassphraseExpiredException
    {
		if (hashSBA.isBurnt()) throw new PassphraseExpiredException();
        return new SensitiveByteArray( hashSBA.getData(), true );
    }
    
	/**
	 * Expire the passphrase.
	 */
	public synchronized void expire()
	{
        hashSBA.burn();                         // burn the data
		if (thread != null) thread.interrupt(); // terminate the thread
	}
	
	public synchronized int hashCode()
	{
        return hashSBA.hashCode();
	}
	
	public synchronized boolean equals( Object o )
	{
        if (o == this) return true;
        if (!(o instanceof Passphrase)) return false;
		Passphrase p = (Passphrase) o;
        synchronized (hashSBA)
        {
		    try
		    {
                SensitiveByteArray pSBA = p.hash();
                if (hashSBA.isBurnt()) return false;
			    return hashSBA.equals( pSBA );
		    }
		    catch (PassphraseExpiredException e)
		    {
			    return !hashSBA.isBurnt();
		    }
        }
	}

	protected void finalize() throws Throwable
	{
		expire();
		super.finalize();
    }

}


/*package-visibility*/ class PassphraseThread extends Thread
{
    private SensitiveByteArray hashSBA;
    private long timeout;
	private boolean die = true;

    // Constructor - called by Passphrase
	PassphraseThread( SensitiveByteArray hashSBA, long timeout )
	{
		super( "Passphrase Thread" );
		this.hashSBA = hashSBA;
		this.timeout = timeout;
		setPriority( getPriority() + 1 );
        setDaemon( true );
		start();
	}

    // Thread execution path - starts running when start() is called (above)
	public void run()
	{
		for (;;)
		{
			try
			{
				sleep( timeout );
				hashSBA.burn();
				return;
			}
			catch (InterruptedException e)
			{
				hashSBA.burn();
				synchronized (this)
				{
                    // if we were interrupted externally then end the thread
					if (die) return;

                    // if we were interrupted by touch() then just loop again
					die = true;
				}
			}
		}
	}

    // touch - called by Passphrase.touch()
    void touch()
	{
		synchronized (this)
		{
			die = false;
			interrupt(); // interrupts run() above
		}
		yield();
	}
	
}

