// This file is currently unlocked (change this line if you lock the file)
//
// Source code version is 2.3.0
//                        (increment this number EVERY time you make a change)
//
// History: (add to this EVERY time you make a change) 
// (insert details of your change immediately below this line)
//
// 2.3.0        16 May 1997     Jill Baker
//      Original Version
//
// If this file is unlocked, or locked by someone other than you, then you
// may download this file for incorporation into your build, however any
// changes you make to your copy of this file will be considered volatile,
// and will not be incorporated into the master build which is maintained
// at this web site.
//
// If you wish to make permanent changes to this file, you must wait until
// it becomes unlocked (if it is not unlocked already), and then lock it
// yourself. Whenever the file is locked by you, you may upload new master
// versions to the master site. Make sure you update the history information
// above. When your changes are complete, you should relinquish the lock to
// make it available for other developers.

/*
 * Copyright (C) 1995, 1996 Systemics Ltd (http://www.systemics.com/)
 * All rights reserved.
 *
 * This library and applications are FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 * as long as the conditions within the COPYRIGHT file are adhered to.
 *
 */

package cryptix.pgp.engine;

import cryptix.io.Lang;
import cryptix.pgp.*;
import cryptix.pgp.engine.toolkit.*;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.PushbackInputStream;
import java.util.Vector;

/**
 * @author Jill Baker
 */
public class DecryptionEngine extends Engine
{
    protected PacketByteArray pba;

    /**
     * Create a DecryptionEngine given a PacketByteArray
     * @param pba PGP packets to be decrypted
     */
    public DecryptionEngine( PacketByteArray pba )
    {
        this.pba = pba;
    }

    /**
     * Return the underlying PacketByteArray
     * @return the underlying PacketByteArray
     */
    public PacketByteArray toPacketByteArray() throws IOException
    {
        return pba;
    }

    /**
     * Completely decrypt a PGP packet stream
     * @param ps Source of PGP packets to be decrypted
     * @param toolkit toolkit containing decryption parameters
     */
    public static void decrypt( PacketInputStream ps, Toolkit toolkit )
    throws IOException, DecryptException
    {
        DecryptStep ds = decryptStep( ps, toolkit );    // decrypt first step of first engine
        if (ds != null)
        {
            if (ds.pba != null) decrypt( ds.pba, toolkit ); // decrypt rest of engine
            if (ds.ps != null) decrypt( ds.ps, toolkit );   // decrypt remaining engines
        }
    }

    /**
     * Completely decrypt an ASCII armoured message
     * @param armouredBytes byte array containing message to be decrypted
     * @param toolkit toolkit containing decryption parameters
     */
    public static void decrypt( byte[] armouredBytes, Toolkit toolkit )
    throws IOException, DecryptException
    {
        // remove armour
        ArmourEngine ae = new ArmourEngine( armouredBytes, toolkit );
        DecryptStep ds = ae.decryptStep( toolkit );     // determine cleartext
        decrypt( ds.pba, toolkit );                     // decrypt rest of engine
    }

    /**
     * Completely decrypt an input stream
     * @param in Source of PGP packets <i>or</i> ASCII armoured input to be decrypted
     * @param toolkit toolkit containing decryption parameters
     * @param maybePGP true if stream contains <i>either</i> PGP packets <i>or</i> ASCII armour;
     *    false if stream definitately contains ASCII armour.
     *    (See also decrypt( PacketInputStream, Toolkit )).
     */
    public static void decrypt( InputStream in, Toolkit toolkit, boolean maybePGP )
    throws IOException, DecryptException
    {
        // Is this an ASCII armoured file or a PGP packet file?
        PushbackInputStream pbs = new PushbackInputStream( in );
        int firstChar = pbs.read();
        pbs.unread( firstChar );

        if ((!maybePGP) || ((firstChar & 0x80) == 0)) // if ASCII armoured file
        {
            // Read whole file into byte array
            ByteArrayOutputStream bs = new ByteArrayOutputStream();
            for (int c=pbs.read(); c!=-1; c=pbs.read()) bs.write( c );
            byte[] armouredBytes = bs.toByteArray();

            // and decrypt it!
            DecryptionEngine.decrypt( armouredBytes, toolkit );
        }
        else // if PGP packet file
        {
            // Just decrypt it!
            DecryptionEngine.decrypt( (new PacketInputStream( pbs )), toolkit );
        }
    }

    /**
     * Completely decrypt a PacketByteArray
     * @param pba PacketByteArray to be decrypted
     * @param toolkit toolkit containing decryption parameters
     */
    public static void decrypt( PacketByteArray pba, Toolkit toolkit )
    throws IOException, DecryptException
    {
        if (pba == null) return;
        decrypt( pba.toPacketInputStream(), toolkit );
    }

    /**
     * Execute a single step in the decryption process
     * @param ps PacketInputStream to be partially decrypted
     * @param toolkit toolkit containing decryption parameters
     * @return DecryptStep containing PacketByteArray and PacketInputStream still to be decrypted
     */
    public static DecryptStep decryptStep( PacketInputStream ps, Toolkit toolkit )
    throws IOException, DecryptException
    {
        DecryptStep ds = new DecryptStep();
        if (ps == null) return ds;
        Packet p = ps.readPacket();
        if (p == null) return ds;
        Engine engine = null;

        switch( p.getType() )
        {
        case Packet.PLAINTEXT:
            {
                engine = new LiteralDataEngine( (LiteralData)p );
                break;
            }
        case Packet.COMPRESSED:
            {
                engine = new CompressionEngine( (CompressedData)p );
                break;
            }
        case Packet.CONV_ENCRYPTED:
            {
                engine = new ConventionalEncryptionEngine( (ConvEncryptedData)p );
                break;
            }
        case Packet.SIGNATURE:
            {
                PacketByteArray pba = new PacketByteArray( ps );
                ps = null;
                int numPackets = pba.numPackets();
                if (numPackets == 0)
                {
                    pba = null;
                    if (toolkit.plaintextGetter == null)
                        toolkit.plaintextGetter = new DetachedPlaintextGetter( toolkit );
                }
                else if ((numPackets == 1) && (pba.getPacketType(0) == Packet.PLAINTEXT))
                {
                    ps = pba.toPacketInputStream();
                    LiteralData ld = (LiteralData) ps.readPacket();
                    if (toolkit.plaintextGetter == null)
                        toolkit.plaintextGetter = new SimplePlaintextGetter( ld.getBuffer() );
                }
                else
                {
                    if (toolkit.plaintextGetter == null)
                        toolkit.plaintextGetter = new SimplePlaintextGetter( pba.toByteArray() );
                }
                engine = new SignatureEngine( (Signature)p, pba );
                break;
            }
        case Packet.PUBLIC_KEY_ENCRYPTED:
            {
                Vector v = new Vector();
                while ((p != null) && (p.getType() == Packet.PUBLIC_KEY_ENCRYPTED))
                {
                    v.addElement( p );
                    p = ps.readPacket();
                }
                if ((p == null) || (p.getType() != Packet.CONV_ENCRYPTED))
                {
                    throw new DecryptException( "Invalid Packet Sequence" );
                }
                engine = new PublicKeyEncryptionEngine( v, (ConvEncryptedData)p );
                break;
            }
        }

        ds = engine.decryptStep( toolkit );
        if (ds == null) ds = new DecryptStep();
        ds.ps = ps;
        return ds;
    }

    /**
     * Execute a single step in the decryption process
     * @param pba PacketByteArray to be partially decrypted
     * @param toolkit toolkit containing decryption parameters
     * @return DecryptStep containing PacketByteArray and PacketInputStream still to be decrypted
     */
    public static DecryptStep decryptStep( PacketByteArray pba, Toolkit toolkit )
    throws IOException, DecryptException
    {
        if (pba == null) return new DecryptStep();
        return decryptStep( pba.toPacketInputStream(), toolkit );
    }

    /**
     * Execute a single step in the decryption process
     * @param toolkit toolkit containing decryption parameters
     * @return DecryptStep containing PacketByteArray and PacketInputStream still to be decrypted
     */
    public DecryptStep decryptStep( Toolkit toolkit ) throws DecryptException, IOException
    {
        return DecryptionEngine.decryptStep( pba, toolkit );
    }

    /**
     * A real (though very simple) command line PGP decryptor!
     */
    public static void main( String[] arg ) throws IOException, DecryptException
    {
        // Uncomment the following line to enable debugging.
        // Utility.debug = new PrintWriter( System.out, true );

        // Gather all the details together
        Toolkit toolkit = new Toolkit( arg[0] );

        // Figure out the destination
        if ((arg.length != 1) && (arg[1].equals( "-p" )))
            toolkit.useStoredFilename = true;
        else
            toolkit.setOutputStream( null, (arg.length==1) ? null : arg[1] );

        // And decrypt...
        DecryptionEngine.decrypt( toolkit.inputStream, toolkit, true );
    }
}
