// This file is currently locked by iang
//
// Source code version is 1.0.0
//                        (increment this number EVERY time you make a change)
//
// History: (add to this EVERY time you make a change) 
// (insert details of your change immediately below this line)
//
// 1.0.0        09 May 1997     Systemics
//      Original version
//

/*
 * Copyright (c) 1995, 1996, 1997 Systemics Ltd
 * on behalf of the Cryptix Development Team.  All rights reserved.
 */
 
package cryptix.security;

import java.io.PrintStream;

/**
 * Implements a Triple DES encryption using the Encrypt-Decrypt-Encrypt
 * (EDE) cycle.
 * Includes the 2-key variant and 3 key variant.
 *
 * <P>
 * TDES takes the DES block cipher and constructs a triple cipher.
 * It thus results in another block cipher in ECB mode.
 * Especially, the use of this by a stream cipher is then possible;
 * this results in <i>outer-feedback</i> mode.
 *
 * <P>
 * This method has been made popular by efforts to strengthen DES, and
 * is often called T-DES or 3-DES.
 *
 * <P>
 * <b>References</b>
 *
 * <P>
 * See also:
 * Schneier, B., "Chapter 15 Combining Block Ciphers,"
 * <cite>Applied Cryptography</cite>, Wiley, 1996, 2nd Ed.<br>
 * ISO 8372, Information Processing Systems: Data Encipherment:
 * Modes of Operation of a 64-bit Block Cipher.<br>
 * ANSI X9.17-1985, American National Standard, Financial
 * Institution Key Management (Wholesale), American Bankers
 * Association, April 4, 1985, Section 7.2.
 *
 * <P>
 * <bold>Copyright</bold> &#169 1995, 1996, 1997
 * <a href="http://www.systemics.com/">Systemics Ltd</a>
 * on behalf of the
 * <a href="http://www.systemics.com/docs/cryptix/">
 * Cryptix Development Team</a>.
 * All rights reserved.
 *
 * <p>
 * @author Systemics Ltd
 * @see BlockCipher
 */
public class TDES extends BlockCipher
{
	private BlockCipher cipher1;
	private BlockCipher cipher2;
	private BlockCipher cipher3;

	//
	// Is this the way to do it?  Or methods?
	//
	static public int MAX_KEY_LENGTH = 24;
	static public int MIN_KEY_LENGTH = 16;
	static public int KEY_LENGTH = 16;

	private int TRIPLE_DES_KEY_LENGTH = 24; // best is 3 * 8 bytes
	private int DES_KEY_LENGTH = 8;	        // manifest
	private int TRIPLE_DES_ROUNDS = 3;      // manifest.  could be 5!
	private int key_length;                 // what we end up using

	/**
	 * The keylength is returned.
	 */
	public int keyLength()
	{
		return key_length;
	}
	public int blockLength()
	{
		return 8;
	}

	/**
	 * Turn a straight DES block cipher into a 3-key triple cipher.
	 * Enciphers and deciphers such:
	 * <pre>
	 *     C = E<sub>k3</sub>(D<sub>k2</sub>(E<sub>k1</sub>(P)))
	 *     P = D<sub>k1</sub>(E<sub>k2</sub>(D<sub>k3</sub>(P)))
	 * </pre>
	 * Keys must be the same length as the cipher's blockLength.
	 *
	 * @param key1    the first key
	 * @param key2    the second key
	 * @param key3    the third key
	 */
	public TDES( byte key1[], byte key2[], byte key3[] )
	{
		set(key1, key2, key3);
	}

	private void set( byte key1[], byte key2[], byte key3[] )
	{
		cipher1 = new DES (key1);
		cipher2 = new DES (key2);
		cipher3 = new DES (key3);

		key_length = 3 * DES_KEY_LENGTH;
	}

	/**
	 * Turn a straight DES block cipher into a 2-key triple cipher.
	 * Enciphers and deciphers such:
	 * <pre>
	 *     C = E<sub>k1</sub>(D<sub>k2</sub>(E<sub>k1</sub>(P)))
	 *     P = D<sub>k1</sub>(E<sub>k2</sub>(D<sub>k1</sub>(P)))
	 * </pre>
	 * Keys must be the same length as the cipher's blockLength.
	 *
	 * @param key1    the first (and third) key
	 * @param key2    the second (or middle) key
	 */
	public TDES( byte key1[], byte key2[] )
	{
		set(key1, key2);
	}

	private void set( byte key1[], byte key2[] )
	{
		// cipher1 = cipher3 = new DES (key1);
		cipher1 = new DES (key1);  // separate for dump purposes for now
		cipher3 = new DES (key1);
		cipher2 = new DES (key2);

		key_length = 2 * DES_KEY_LENGTH;
	}


	/**
	 * Make a triple cipher out of straight DES.
	 * 2-key mode is used if the key is 16 bytes, 3-key if it
	 * is 24 bytes long.
	 *
	 * @param key     the key is either 2 or 3 times a keyLength
	 */
	public TDES( byte key[] )
		throws CryptoError
	{
		int length = key.length;
		int single = DES_KEY_LENGTH;

		byte one[]   = new byte [single];
		byte two[]   = new byte [single];
		byte three[] = new byte [single];

		if (24 == length)
		{
			System.arraycopy(key, 0,        one,   0, single);
			System.arraycopy(key, single,   two,   0, single);
			System.arraycopy(key, 2*single, three, 0, single);

			// System.out.println("KEYS ARE [ " +
			// 			toString(one) + " ] [ " +
			// 			toString(two) + " ] [ " +
			// 			toString(three) + " ]");
			set(one, two, three);
		}
		else if (16 == length)
		{
			System.arraycopy(key, 0,        one,   0, single);
			System.arraycopy(key, single,   two,   0, single);

			// System.out.println("KEYS ARE [ " +
			// 			toString(one) + " ] [ " +
			// 			toString(two) + " ]");


			set(one, two);
		}
		else
			throw new CryptoError( "TDES: " + length +
				" illegal length for triple DES key" +
				" (16 or 24 bytes only)" );
	}


	/**
	 * This function encrypts a block of data.
	 * The contents of the array in remain unchanged,
	 * with the result stored in the array out.
	 * The arrays can, however, be the same.
	 *
	 * @param in          The plain text to be encrypted.
	 * @param in_offset   The offset within the in buffer.
	 * @param out         Where the encrypted cipher text will be stored.
	 * @param out_offset  The offset within the out buffer.
	 */
	protected void 
	blockEncrypt( byte in[], int in_offset, byte out[], int out_offset )
	{
		byte[] tmp = new byte [DES_KEY_LENGTH];

		// System.out.println("    READY       " + toString(in));
		cipher3.encrypt(in, in_offset, tmp, 0);
		// System.out.println("    AFTER ONE   " + toString(tmp));
		cipher2.decrypt(tmp);
		// System.out.println("    AFTER TWO   " + toString(tmp));
		cipher1.encrypt(tmp, 0, out, out_offset);
		// System.out.println("    AFTER THREE " + toString(out));
	}

	/**
	 * This function decrypts a block of data.
	 * The contents of the array in will not be changed,
	 * but will instead store the result in the array out.
	 * The arrays can, however, be the same.
	 *
	 * @param in          The cipher text to be decrypted.
	 * @param in_offset   The offset within the in buffer.
	 * @param out         Where the decrypted plain text will be stored.
	 * @param out_offset  The offset within the out buffer.
	 */
	protected void
	blockDecrypt( byte in[], int in_offset, byte out[], int out_offset )
	{
		byte[] tmp = new byte [DES_KEY_LENGTH];

		cipher1.decrypt(in, in_offset, tmp, 0);
		cipher2.encrypt(tmp);
		cipher3.decrypt(tmp, 0, out, out_offset);
	}

	public String dump( int depth )
	{
		if (depth <= 0) return "";

		String s = new String("TDES: dump");

		s += " " + keyLength() + "/" + blockLength() + "/" + TRIPLE_DES_ROUNDS;

		if (--depth <= 0) return s ;

		s += "\nONE   " + cipher1.dump(depth);
		s += "\nTWO   " + cipher2.dump(depth);
		s += "\nTHREE " + cipher3.dump(depth);

		return s ;
	}
///////////////////////////////// T E S T ////////////////////////

/**
 * Entry point for <code>self_test</code>.
 */
public static final void
main(String argv[]) 
{
	try { self_test(); }
	catch(Throwable t) { t.printStackTrace(); }
}


//
// I have no TDES test data - I made this up myself.
// Use ints as Java grumbles about negative hex values.
//
static final private byte[][][] tests =
{
  {
    { 1, 35, 69, 103, -119, -85, -51, -17, 85, 85, 85, 85, 85, 85, 85, 85, },
    { 115, 111, 109, 101, 100, 97, 116, 97, },   // plain
    { -5, -85, -95, -1, -99, 5, -23, -79, },     // cipher
  },
  {
    {
      1, 35, 69, 103, -119, -85, -51, -17, 85, 85, 85,
      85, 85, 85, 85, 85, -2, -36, -70, -104, 118, 84, 50, 16,
    },
    { 115, 111, 109, 101, 100, 97, 116, 97, },
    { 5, -3, -71, -127, -34, 99, -62, 22, },
  },
  {
    { // key
      3, 82, 2, 7, 103, 32, -126, 23, -122, 2, -121, 102, 89,
      8, 33, -104, 100, 5, 106, -67, -2, -87, 52, 87,
    },
    {  115, 113, 117, 105, 103, 103, 108, 101, }, // plain
    { 82, 30, -28, -30, 72, -113, 91, 90, },      // cipher
  },
};

/**
 * Do some basic tests.
 * Three of the certification data are included only, no output,
 * success or exception.
 * If you want more, write a test program!
 * @see cryptix.examples.TestTDES
 */
public static final void
self_test()
	throws CryptoError
{
	for (int i = 0; i < tests.length; i++)
	{
		byte[][] test = tests[i];
		byte[] key    = test[0];
		byte[] plain  = test[1];
		byte[] cipher = test[2];

		TDES cryptor = new TDES(key);
		byte[] tmp = new byte[cryptor.blockLength()];

		cryptor.encrypt(plain, tmp);
		if (notEquals(cipher, tmp))
			throw new CryptoError("encrypt #"+ i +" failed");
		cryptor.decrypt(cipher, tmp);
		if (notEquals(plain, tmp))
			throw new CryptoError("decrypt #"+ i +" failed");
	}
}

private static final boolean
notEquals(byte[] a, byte[] b)
{
	for (int i = 0; i < a.length; i++)
	{
		if (a[i] != b[i])
			return true;
		// if (a[i] == 90) // force a trip up
		// 	return true;
	}
	return false;
}

} /* class */
