package enigma;

import java.io.IOException;
import java.io.FileNotFoundException;
import systemics.util.IniFile;
import systemics.util.IniFileFormatException;

/**
 * Provides a cleaner, application-specific 
 * interface to the Enigma preferences file.
 *
 * Should allow transfer to Java 1.1's newer 
 * registry functions relatively painlessly.
 *
 * @author Ian Brown
 * @version 0.9
 */
class Preferences extends IniFile
{
	// Critical options - MUST be set in preferences file
	String myName_ = null;

	// Non-critical options with defaults
	boolean acceptRemoteClients_ = false;

	public Preferences( String filename )
	throws IOException, FileNotFoundException, IniFileFormatException
	{
		super( filename );

		// Initialise critical options
		// If any of these are missing, an exception will be thrown
		myName_ = get( "General", "MyName" );
		if ( myName_ == null ) throw new IniFileFormatException( "You MUST include the MyName option in your preferences file." );

		// Load non-critical options
		if ( getBool( "Security", "AcceptRemoteClients" ) )
			acceptRemoteClients_ = true;

	}

	public final String myName()
	{
		return myName_;
	}

	public final boolean acceptRemoteClients()
	{
		return acceptRemoteClients_;
	}
	
	public final int verbosity()
	{
		int level = getInt( "General", "Verbosity", 0 );
		switch ( level )
		{
			case Plugin.QUIET: return Plugin.QUIET;
			case Plugin.VERBOSE: return Plugin.VERBOSE;
			default: return Plugin.NORMAL;
		}
	}
		
	// General functions for plug-ins
	public final int getInt( String section, String option )
	throws NumberFormatException
	{
		String value = get( section, option );
		if ( value == null ) return -1; // ints can't be null so use -1 instead
		else return Integer.valueOf( value ).intValue();
	}

	public final int getInt ( String section, String option, int defaultValue )
	{
		int value = getInt( section, option );
		if ( value  == -1 ) return defaultValue;
		else return value;
	}
	
	/**
	 * Default for this function is false
	 * True will only be returned if the value of option 
	 * is yes, true or -1.
	 */
	public final boolean getBool( String section, String option )
	{
		// Deal with any combination of upper or lower case
		String value = get( section, option );
		if ( value == null ) return false;
		else value = value.toLowerCase();
		if ( value.equals( "yes" ) || value.equals( "true" ) || value.equals( "-1" ) )
			return true;
		else return false;
	}
}
