package enigma;

import java.io.*;
import java.net.*;

import java.util.Vector;
import java.util.StringTokenizer;

import cryptix.security.rsa.PublicKey;
import cryptix.security.rsa.SecretKey;
import cryptix.pgp.*;

import cryptix.util.EmailAddress;
import systemics.util.IniFileFormatException;

/**
 * Plugin to handle SMTP connections.
 * @author Ian Brown
 * @version 0.93 - option not to clearsign messages
 * Version 0.92 - closed connections after an error.
 * Version 0.91 - second beta. Fixed trailing whitespace problem.
 * Version 0.9 - first beta
 */ 
class SMTP extends Plugin
{
	private static final int DEFAULT_SMTP_SERVER_PORT = 25;
	private static final int DEFAULT_SMTP_MY_PORT = 25;
	
	// Sockets and streams for communicating with client and server
	ServerSocket server = null;
      Socket clientSocket = null;
      Socket serverSocket = null;

      DataInputStream clientIn = null;
      DataInputStream serverIn = null;
      PrintStream clientOut = null;
      PrintStream serverOut = null;

      String command = null;
      String reply = null;      
        
      // SMTP Reply codes
      static final String SMTP_OK = "250";
      static final String SMTP_MULTILINE_HELP = "214";
      static final String SMTP_CLOSING = "21";
      static final String SMTP_INTERMEDIATE_OK = "354";
        
	Vector to = new Vector(); // List of public keys of recipients
	
	boolean clearsignOnly = false;
	// Change below to true if you want to clearsign again
	boolean clearsign = false;	
	
	public SMTP( KeyGlob globby, Passphrase pphrase, Preferences prefs0 )
	throws IOException, IniFileFormatException, Exception
	{          
    		super( globby, pphrase, prefs0 );
    		// If options have not been set in preferences file, use defaults
    		if ( serverPort == -1 ) serverPort = DEFAULT_SMTP_SERVER_PORT;
    		if ( myPort == -1 ) myPort = DEFAULT_SMTP_MY_PORT;
                clearsign = prefs.getBool("General","ClearSign");
	}

	public void run()
	{
		try
		{
       		// Create server object
			server = new ServerSocket( myPort );
		}
	
		catch ( IOException ioe )
		{
     			if ( verbosity >= NORMAL )
    			{
    				System.err.println( "Couldn't create a TCP socket on port " + myPort );
    				if ( myPort < 1025 ) System.err.println( "You need system administration privileges to use a port lower than 1025." );
    				else System.err.println( "Is another program using it?" );
    				System.err.println( "Try changing the line MyPort = " + myPort + " in the [SMTP] section of your " + prefs.name() + " file." );
    			}
    			
    			System.exit( PLUGIN_PROBLEM );
		}	
	
		while ( true )
		{
			try
			{
				EmailAddress sender;

				// Wait for client to connect
        			// then create socket
        			clientSocket = server.accept();
				if ( !InetAddress.getLocalHost().equals( clientSocket.getInetAddress() ) && !prefs.acceptRemoteClients() )
					throw new RemoteAccessAttemptException();
				
        			clientIn = new DataInputStream( clientSocket.getInputStream() );
        			clientOut = new PrintStream( clientSocket.getOutputStream(), true );  // true autoflushes stream after newlines
    
        			// Open connection to true mail server
        			try
        			{
            			serverSocket = new Socket( InetAddress.getByName( serverName ), serverPort );
        			}

        			catch( UnknownHostException e )
        			{
        				if ( verbosity >= NORMAL )
    					{   						
            					System.err.println( "Can't find SMTP server " + serverName );
            					System.err.println( "Check that your configuration file " + prefs.name() + " contains the correct Internet name." );
        				}
        			}

        			serverIn = new DataInputStream( serverSocket.getInputStream() );
        			serverOut = new PrintStream( serverSocket.getOutputStream(), true );

        			// Get greeting from server
        			reply = serverIn.readLine();
        			if ( verbosity >= VERBOSE ) System.out.println( reply );
        			clientOut.println( reply );

        			while ( !reply.substring(1,3).equals( SMTP_CLOSING ) )
        			// Reply codes 421 and 221 imply connection is about to be closed.
        			// No other codes end in 21
        			{
            			// Get command and make it upper-case (SMTP is case-insensitive)
            			command = clientIn.readLine();
            			command.toUpperCase();
            				           				            
            			serverOut.println(command);

            			// Get reply
           				reply = serverIn.readLine();
            			clientOut.println(reply); // Send reply
           
            			// Deal with command
					if ( command.startsWith( "MAIL FROM:" ) )
					{
						sender = new EmailAddress( command );
						System.out.println( "Message sender: " + sender );
					}

            			else if( command.startsWith( "RCPT" ) )
       	    			{
            				// Get TO: field to fetch public key
            				EmailAddress recipient = new EmailAddress( command );
            				if ( verbosity >= VERBOSE ) System.out.println( "Recipient: " + recipient.toString() );
            				PublicKey toKey = null;
						// If the message doesn't already have a recipient without a public key,
						// try and get a key for this recipient.
            				if ( !clearsignOnly ) toKey = keyGlob.getKey( recipient );
            				if ( toKey == null ) clearsignOnly = true; // If one of the
            				// recipients of a message does not have a public key, the
            				// message will be clearsigned ONLY to ALL recipients.
            				else to.addElement( toKey );
            			}
            
            
            			else if ( command.startsWith( "RSET" ) ) reset();
            
            			else if ( command.startsWith( "EXPN" ) || command.startsWith("HELP" ) )
            			{
                				// Response will be multi-line if +OK
               				if ( reply.startsWith(SMTP_OK) || reply.startsWith( SMTP_MULTILINE_HELP) )
                				{
                    				while( reply != null ) // SMTP doesn't specify a termination signal in this case
                    				{
                    					reply = serverIn.readLine();
                        				clientOut.println(reply);
                    				}
                				}
           				}
            
           				else if ( command.equals("DATA") )
            			{
            				SecretKey from = myKey;
            	            	
            				if ( reply.startsWith( SMTP_INTERMEDIATE_OK ) )
            				{
            					StringBuffer msgText = new StringBuffer(5096);  // Stores message body

							// Message will be signed, so get user's passphrase/private key here?
							// or should ArmouredMessage/SecretKeyRing do the job for itself?
							// Probably, to manage the hashtable of passphrases.
							
							// Get and send on message header
            					while (command.length() != 0) // ie is not a blank line
            					{
            						// Send header unencrypted so non PGP-enabled mailers can understand it
            						command = clientIn.readLine();          			
            						serverOut.println( command );
            					}

            		            	command = clientIn.readLine(); // Get first line of message body
	
							if (clearsignOnly && !clearsign)
							{
								// We don't need to sign OR encrypt message so just pass straight through
								do {
									serverOut.println(command);
									command = clientIn.readLine(); 
								} while(!command.equals( "." ));

							}         

            					else {
								if (clearsignOnly) to = null;
								
            						do {
            							command.trim(); // PGP is very fussy about trailing whitespace
									// but this may cause problems as trim() cuts off starting whitespace as well
            							msgText.append(command).append(Armoury.LINEFEED);
            							command = clientIn.readLine();         		
            						} while(!command.equals("." ));

            						// Get some timing information
                        				long startTime = System.currentTimeMillis();		
            						ArmouredMessage encryptedMessage = new ArmouredMessage( msgText.toString(), from, to );           						
            						long elapsedTime = System.currentTimeMillis() - startTime;           						
            						serverOut.print( encryptedMessage.ciphertext() );
            						
            						if ( verbosity >= VERBOSE )
            						{
            							System.out.println( encryptedMessage.ciphertext() );
            							System.out.println( "[SMTP] That took " + elapsedTime + " milliseconds." );
            						}
							}

            					serverOut.println( "." );  // Send message terminator

            					// Get server's response
            					reply = serverIn.readLine();
            					clientOut.println(reply);
            				}
            			}                                
				}
			}
      
			catch ( NullPointerException npe )
			{
      			// Some badly-behaved SMTP clients don't close the connection to the server correctly
      			// by sending a QUIT message - they just close the socket. This causes a NullPointerException,
      			// which this catch block deals with by simply restarting the plugin.
				if ( verbosity >= VERBOSE ) System.err.println( "Mail client or SMTP server broke its connection." );
				npe.printStackTrace();
				
			}
			
			catch ( RemoteAccessAttemptException raae )
			{
				if ( verbosity >= NORMAL ) 
				{					
					System.err.println( "SECURITY ALERT: someone at Internet address " + clientSocket.getInetAddress() );
					System.err.println( "tried to connect to your SMTP plug-in. This would allow them to sign messages" );
					System.err.println( "as though they came from you. The request was denied." );
				}
			}			
      	
			catch ( Throwable t )
			{	
      				// Problems are 'dealt' with by restarting the loop.
      				// This is done in order to allow this plugin to be a thread; to extend Thread you have
      				// to deal with all errors yourself instead of being able to throw some of them to the Java VM.
      				if ( verbosity >= NORMAL ) t.printStackTrace();

					// Close any still-open connections
					try
					{
						serverSocket.close();
						clientSocket.close();
					}
				
					catch ( IOException ioe )
					{
					}
			}
      
			// Well-behaved mailers will also get to this point, having properly closed the socket
			reset();
		}
	} // End run()
    
	public void reset()
	{
    		clearsignOnly = false;
    		to = new Vector();
	}
	
	public String name()
	{
		return "SMTP";
	}

} // End class SMTP
