package enigma;

import java.io.*;
import java.awt.Frame;
import java.awt.event.*;
import java.util.StringTokenizer;
import java.util.Properties;
import java.security.SecureRandom;
import cryptix.security.rsa.SecretKey;

import cryptix.pgp.*;

import systemics.util.IniFileFormatException;

/**
 * Initialises your keyrings and the preferences file specified on the
 * command line then loads all your protocol handlers.
 * @author Ian Brown
 * @version 0.91 - asked for passphrase for specific ID.
 */
public class Startup extends Frame implements WindowListener
{	
	// Abnormal exit codes
	private static final int NO_PREFS = 1;
	private static final int PREFS_DAMAGED = 2;
	private static final int PREFS_MISSING = 3;
	private static final int KEYRING_PROBLEM = 4;
	private static final int WRONG_PASSPHRASE = 5;
		
	public Startup( String[] argv )
	throws Exception
	{
		// Only extends Frame because Dialog (which PassphraseDialog class uses
		// to get the user's passphrase) requires a 'parent' Frame.
		// This class is never show()n so it is invisible.
		super("Enigma");
		this.addWindowListener(this);

		SecretKeyRing secretKeyRing = null;
		PublicKeyRing publicKeyRing = null;
		KeyGlob keyGlob = new KeyGlob();
		SecretKey myKey = null;
	
		Passphrase passphrase;
		Preferences myPrefs = null;
		
    		// Load preferences from .ini file on command line
    		// Assume normal verbosity - we can't find out the
    		// user setting until the prefs file is loaded!	
    		
		if ( argv.length < 1 )
		{
			System.err.println( "Usage: java enigma.Startup inifilename" );
			System.err.println( "Enigma MUST be started with a preferences file." );
			System.exit( NO_PREFS );
		}

		try
		{
			myPrefs = new Preferences( argv[0] );
		}

		catch ( IniFileFormatException e )
		{
			System.err.println( "Your preferences file " + argv[0] + " is damaged," );
			System.err.println( "or is missing important information:" );
  			System.err.println( e.getMessage() );
			System.exit( PREFS_DAMAGED );
		}
	
		catch ( FileNotFoundException e )
		{
			System.err.println( "Couldn't find preferences file " + argv[0] );
			System.exit( PREFS_MISSING );
		}
		
	
		// Get passphrase from user
		// Have to DIY for the moment as cryptix.gui classes aren't finished yet
		PassphraseDialog dialog = new PassphraseDialog( this, "Enigma", myPrefs.get( "General", "MyName") ); // Problem line!
		String phrase = dialog.getPassphrase();
		passphrase = new Passphrase( phrase );
		
		// Try and wipe plaintext passphrase from memory
		dialog = null;
		phrase = null;
		System.gc();

		// Get keyrings
		try
		{
			StringTokenizer filenames;
			
			// Add list of space-separated public and secret keyrings to keyGlob
			filenames = new StringTokenizer( myPrefs.get( "KeyRings", "SecretKeyRing" ) );
			while ( filenames.hasMoreTokens() )
			{
				secretKeyRing = new SecretKeyRing( filenames.nextToken() );
				keyGlob.addKeyRing( secretKeyRing );
			}

			filenames = new StringTokenizer( myPrefs.get( "KeyRings", "PublicKeyRing" ) );
			while ( filenames.hasMoreTokens() )
			{
				publicKeyRing = new PublicKeyRing( filenames.nextToken() );
				keyGlob.addKeyRing( publicKeyRing );
			}
		}
		
		catch ( Exception e )
		{
			if ( myPrefs.verbosity() >= Plugin.NORMAL )
			{
				System.err.println( "Couldn't load one of the keyring files specified in your preferences file." );
				System.err.println( e.getMessage() );
			}
			
			System.exit( KEYRING_PROBLEM );
		}
		
		// Set the keyserver to use
		if ( myPrefs.getBool( "KeyServer", "UseWebKeyServer" ) )
		{
			// Use MIT's fast keyserver by default
			String host = myPrefs.get( "WebClient", "Server", "swissnet.ai.mit.edu" );
			int port = myPrefs.getInt( "WebClient", "Port", 11371 );
			String path = myPrefs.get( "WebClient", "Path", "/pks/lookup" );

			WebClient client = new WebClient ( host, port, path );

			/* not currently right
			// Now add settings for caching and exclusion list, if set
			String cacheName = myPrefs.get( "WebClient", "CacheFile" );
			if ( cacheName != null )
			{
				System.out.println( "Using user-specified cache" );
				KeyCache cache = null;
				try
				{
					cache = new KeyCache( cacheName );
				}
				
				catch ( Exception e )
				{
					// Just use a temporary cache
					cache = new KeyCache();
				}
									
				String excludeFile = myPrefs.get( "WebClient", "ExcludeFile" );
				if ( excludeFile != null )
				{
					try
					{
						cache.setExclusionFile( new File( excludeFile ) );
					}
					
					catch ( FileNotFoundException fnfe )
					{
						if ( myPrefs.verbosity() >= Plugin.NORMAL )
						{
							// This may be the first time the file is being used
							System.err.println( "Couldn't find the file you specified as an exclusion list." );
							System.err.println( "Creating " + excludeFile + " now.");
						}
						
						FileOutputStream fos = new FileOutputStream( excludeFile );
						fos.close();
						cache.setExclusionFile( new File( excludeFile ) );
					}
					
					catch ( Exception e )
					{
						// Just use a temporary list
						if ( myPrefs.verbosity() >= Plugin.NORMAL )
						{
							System.err.println( "Problem loading the exclusion list specified in your initialisation file:" );
							System.err.println( e.getMessage() );
						}
						cache.setExclusionList( new Properties() );
					}
				}
				client.setCache( cache );
			} */
			keyGlob.setKeyServer( client );
		}
		
		else if ( myPrefs.verbosity() >= Plugin.VERBOSE ) System.out.println( "Not using keyserver" );
	
		try
		{
				myKey = keyGlob.getSecretKey( myPrefs.get( "General", "MyName"), passphrase );
		}
 
 		catch ( DecryptException e )
 		{
			if ( myPrefs.verbosity() >= Plugin.NORMAL )
			{
				System.err.println( "Couldn't decrypt secret key for " +  myPrefs.get( "General", "MyName" ) );
 				System.err.println( "Passphrase may be incorrect." );
 			}
 			
 			System.exit( WRONG_PASSPHRASE );
 		}
 
 		if ( myKey == null ) throw new Exception( "Couldn't load your secret key." );	
	
	
		// Load protocol handlers
		SMTP SMTPproxy = new SMTP( keyGlob, passphrase, myPrefs );
		POP3 POPproxy = new POP3( keyGlob, passphrase, myPrefs );
		
		// Set them off as multitasking daemon services
		SMTPproxy.start();
		POPproxy.start();
		
		if ( myPrefs.verbosity() >= Plugin.VERBOSE ) System.out.println( "Threads started..." );

		// This isn't used by Startup but it starts Java calculating a random seed
		// which will then be available when SMTP needs it. (Java maintains a VM-wide
		// seed)
		SecureRandom initialiseRandomSeed = new SecureRandom();

		//show();
		SMTPproxy.join(); // Don't want to quit now as the windowClosing event would be called
	}

	public void windowClosing(WindowEvent e) { System.exit(0); }
	public void windowOpened(WindowEvent e) {}
	public void windowClosed(WindowEvent e) {}
	public void windowIconified(WindowEvent e) {}
	public void windowDeiconified(WindowEvent e) {}
	public void windowActivated(WindowEvent e) {}
	public void windowDeactivated(WindowEvent e) {}

	public static void main (String argv[])
	throws Exception
	{
		Startup go = new Startup( argv );
	}
}
